from visual import *
''' space for constants, importing things, or other documentation '''

class RubiksCube:
    ''' keeps a list of all the cubelets, locations in array based on initial positions '''
    cubeletList = [[[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]]]

    ''' keeps a list of all the cubelets, locations in array based on current positions '''
    rCState = [[[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]]]

    ''' maintains a list of moves made so that it can "solve" the rubiks cube by reversing them '''
    movesMade = []
    
    def __init__(self):
        self.initRubiksCube()
        cubeletList = [[[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]]]
        rCState = [[[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]],
                    [[None, None, None],
                    [None, None, None],
                    [None, None, None]]]
        movesMade = []

    def initRubiksCube(self):
        for X in range(-1, 2):
            for Y in range(-1, 2):
                for Z in range(-1, 2):
                    pos = (1.15*X, 1.15*Y, 1.15*Z) #1.15 puts .15 distance between each cubelet
                    faces = [False, False, False, False, False, False] #keeps track of which faces are colored
                                                                        #NOT ORIENTATIONS
                    if(X == 1):
                        faces[0] = True
                    if(X == -1):
                        faces[1] = True
                    if(Y == 1):
                        faces[2] = True
                    if(Y == -1):
                        faces[3] = True
                    if(Z == 1):
                        faces[4] = True
                    if(Z == -1):
                        faces[5] = True
                    cube = Cubelet(pos, faces)
                    self.cubeletList[X+1][Y+1][Z+1] = cube
                    self.rCState[X+1][Y+1][Z+1] = cube

    def resetCube(self):
        print("in resetCube")
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    self.cubeletList [i][j][k].delete()
                    print('deleted', i, j, k)
                    self.cubeletList[i][j][k] = None

    def undo(self):
        if(len(self.movesMade) > 0):
            loc = len(self.movesMade) - 1
            origAxis = self.movesMade[loc][1]
            reverseAxis = (-origAxis[0], -origAxis[1], -origAxis[2])
            cor = calculateCOR(self.movesMade[loc][0], reverseAxis)
            rotateList = self.movesMade[loc][2]
            self.rotateSide(cor, reverseAxis, rotateList, 45)
            self.movesMade.pop(loc)
        
        
    def runOneMove(self):
        keypress1 = scene.waitfor('keydown')
        cubeletPos = (0,0,0)
        v = keypress1.key
        if(v == 'i'):
            self.rotateCamera((-1,0,0))
        elif(v == 'k'):
            self.rotateCamera((1,0,0))
        elif(v == 'j'):
            self.rotateCamera((0,-1,0))
        elif(v == 'l'):
            self.rotateCamera((0,1,0))
        elif(v == 'q'):
            cubeletPos = (-1.15,1.15,1.15)
        elif(v == 'w'):
            cubeletPos = (0,1.15,1.15)
        elif(v == 'e'):
            cubeletPos = (1.15,1.15,1.15)
        elif(v == 'a'):
            cubeletPos = (-1.15,0,1.15)
        elif(v == 's'):
            cubeletPos = (0,0,1.15)
        elif(v == 'd'):
            cubeletPos = (1.15,0,1.15)
        elif(v == 'z'):
            cubeletPos = (-1.15,-1.15,1.15)
        elif(v == 'x'):
            cubeletPos = (0,-1.15,1.15)
        elif(v == 'c'):
            cubeletPos = (1.15,-1.15,1.15)
        elif(v == 'b'):
            self.scramble(random.randint(20,50))
        elif(v == 'r'):
            ''' faulty implementation of resetting the RTICA
                feature not included in final project
                    note that solve() serves basically the same purpose as a reset
            self.resetCube()
            self.initRubiksCube()
            self.normalizePositions()
            self.updateRCState()
        '''
        elif(v == 'o'):
            orbit()
        elif(v == 'f'):
            rateValue = len(self.movesMade)*3
            self.solve(max(rateValue,90))
        elif(v == 'v'):
            self.undo()
        elif(v == 'esc'):
            return False
        if(cubeletPos != (0,0,0)):
            selected = self.findCubelet(cubeletPos)
            selected.highlight()
            aor = (0,0,0)
            while(aor == (0,0,0)):
                keypress2 = scene.waitfor('keydown')
                w = keypress2.key
                if(w == 'q'):
                    aor = (0,0,1)
                elif(w == 'e'):
                    aor = (0,0,-1)
                elif(w == 'w'):
                    aor = (-1,0,0)
                elif(w == 'x'):
                    aor = (1,0,0)
                elif(w == 'a'):
                    aor = (0,-1,0)
                elif(w == 'd'):
                    aor = (0,1,0)
                elif(w == 'esc'):
                    return False
                else:
                    break
            if(aor != (0,0,0)):
                cor = calculateCOR(cubeletPos, aor)
                rotateList = self.generateRotateList(cor, aor)
                self.rotateSide(cor, aor, rotateList, 45)
                self.updateMovesMade(cubeletPos, aor, rotateList)
            selected.unhighlight()
        return True
        

    ''' ensures that the index [i][j][k] in rCState corresponds to the cubelet currently at location
        (1.15*i - 1.15, 1.15*j - 1.15, 1.15*k - 1.15) '''
    def updateRCState(self):
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    self.rCState[i][j][k] = self.findCubelet((1.15*i - 1.15, 1.15*j - 1.15, 1.15*k - 1.15))

    def findCubelet(self, pos):
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    testCubelet = self.cubeletList[i][j][k]
                    if(testCubelet.pos == pos):
                        return testCubelet
        
    def normalizePositions(self):
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    toCheck = vector(self.cubeletList[i][j][k].pos)
                    for v in range(3):
                        if(toCheck[v] < .01 and toCheck[v] > -.01):
                            toCheck[v] = 0
                        elif(toCheck[v] < 1.16 and toCheck[v] > 1.14):
                            toCheck[v] = 1.15
                        elif(toCheck[v] < -1.14 and toCheck[v] > -1.16):
                            toCheck[v] = -1.15
                    self.cubeletList[i][j][k].pos = toCheck

    ''' takes a center of rotation, an axis, and a list and rotates all the cubelets in the list
        at 30 fps. then it normalizes the positions and updates the state of the rubiks cube '''
    def rotateSide(self, cor, axis, rotateList, rateValue):
        for i in range(30):
            rate(rateValue)
            for cubelet in rotateList:
                cubelet.rotateCubeletByAngle(cor, axis, pi/60)
        self.normalizePositions()
        self.updateRCState()

    ''' easier to use rotate method '''
    def rotateWrapper(self, cubeletPos, aor, rateValue):
        cor = calculateCOR(cubeletPos, aor)
        rotateList = self.generateRotateList(cubeletPos, aor)
        self.rotateSide(cor, aor, rotateList, rateValue)
        
    def solve(self, rateValue):
        for i in range(len(self.movesMade)-1, -1, -1):
            reverseAxis = (-self.movesMade[i][1][0],
                           -self.movesMade[i][1][1],
                           -self.movesMade[i][1][2])
            cor = calculateCOR(self.movesMade[i][0], reverseAxis)
            rotateList = self.movesMade[i][2]
            self.rotateSide(cor, reverseAxis, rotateList, rateValue)
        self.movesMade = []
            
    ''' adds whatever move is made to the list of moves made '''
    def updateMovesMade(self, cubeletPos, axis, rotateList):
        self.movesMade.append((cubeletPos, axis, rotateList))

    ''' generates the list of cubelets to be rotated, given a cubelet position and an axis of rotation '''            
    def generateRotateList(self, pos, axis):
        rotateList = []
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    if(axis[2] != 0 and k == 2):
                        rotateList.append(self.rCState[i][j][k])
                    elif(axis[0] != 0 and i*1.15 - 1.15 == pos[0]):
                        rotateList.append(self.rCState[i][j][k])
                    elif(axis[1] != 0 and j*1.15 - 1.15 == pos[1]):
                        rotateList.append(self.rCState[i][j][k])
        return rotateList

    def rotateCamera(self, axis):
        rCList = []
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    rCList.append(self.cubeletList[i][j][k])
        self.rotateSide((0,0,0), axis, rCList, 30)
        self.updateMovesMade((0,0,0), axis, rCList)

    def scramble(self, numMoves):
        possiblePositions = [(-1.15,1.15,1.15), (0,1.15,1.15), (1.15,1.15,1.15),
                             (-1.15,0,1.15), (0,0,1.15), (1.15,0,1.15),
                             (-1.15,-1.15,1.15), (0,-1.15,1.15), (1.15,-1.15,1.15)]
        possibleAxes = [(1,0,0), (-1,0,0), (0,1,0), (0,-1,0), (0,0,1), (0,0,-1)]
        for i in range(numMoves):
            cubeletPos = possiblePositions[random.randint(0,8)]
            aor = possibleAxes[random.randint(0,5)]
            cor = calculateCOR(cubeletPos, aor)
            rotateList = self.generateRotateList(cor, aor)
            self.rotateSide(cor, aor, rotateList, numMoves*10)
            self.updateMovesMade(cubeletPos, aor, rotateList)

class Cubelet:
    axList = [(0,1,0),(-1,0,0),(0,0,-1),(0,0,1),(0,1,0),(0,-1,0)]
    colorList = [color.red, color.magenta, color.blue, color.green, color.white, color.yellow]
    
    def __init__(self, pos, faces):
        self.pos = pos
        self.faces = faces
        self.pyramidList = []
        self.makePyramidList(pos, faces)

    def makePyramidList(self, pos, faces):
        height = 0.5
        baseLength = 1
        pyrPos = (pos[0] - 0.5, pos[1], pos[2]) #0.5 is the height of each pyramid
        for i in range(6):
            cubeletFace = pyramid(pos = pyrPos, size = (.5, 1, 1), color = Cubelet.colorList[i])
            if(not faces[i]):
                cubeletFace.color = color.black
            cubeletFace.rotate(angle = pi/2, axis = Cubelet.axList[i], origin = pos)
            if(i == 0):
                cubeletFace.rotate(angle = pi/2, axis = Cubelet.axList[i], origin = pos)
            self.pyramidList.append(cubeletFace)

    ''' rotates an entire cubelet some given angle, with animations and an internal rate '''
    #this method is not in use
    def rotateCubelet(self, cor, aor):
        for t in range(30): #these two lines mean the rotation takes one second
            rate(30)
            for i in range(6):
                self.pyramidList[i].rotate(angle = pi/60, axis = aor, origin = cor)

    ''' rotates an entire cubelet some angle, almost instantly, without animation '''
    def rotateCubeletByAngle(self, cor, aor, a):
        for i in range(6):
            self.pyramidList[i].rotate(angle = a, axis = aor, origin = cor)
        self.updateCubeletPos()

    ''' ensure's the the cubelet's state value represents its actual state, since it isn't
        autmomatically updated when the pyramids are moved '''
    def updateCubeletPos(self):
        self.pos = average(self.pyramidList[0].pos, self.pyramidList[1].pos)

    def highlight(self):
        for pyr in self.pyramidList: #trying out a little bit cleaner notation here
            pyr.opacity = 0.4

    def unhighlight(self):
        for pyr in self.pyramidList:
            pyr.opacity = 1

    def delete(self):
        '''
        for pyr in self.pyramidList:
            pyr.visible = False
            del pyr
        '''
        for i in range(len(self.pyramidList)):
            self.pyramidList[0].visible = False
            del self.pyramidList[0]

''' space for general utility methods '''

''' averages two positions '''
def average(pos1, pos2):
    return ((pos1[0]+pos2[0])/2, (pos1[1]+pos2[1])/2, (pos1[2]+pos2[2])/2)

'''calculates the center of rotation for a rotation, given the position of a selected cubelet
    and an axis of rotation '''
def calculateCOR(cubeletPos, aor):
    if(aor[2] != 0):
        return (0,0,cubeletPos[2]) #should be (0,0,1.15)
    elif(aor[1] != 0):
        return (0,cubeletPos[1],0)
    elif(aor[0] != 0):
        return (cubeletPos[0],0,0)

''' rotates the camera around the cube slowly '''
def orbit():
    storedForward = (scene.forward[0], scene.forward[1], scene.forward[2])
    cont = True
    while(cont == True):
        rate(30)
        if scene.kb.keys:
            s = scene.kb.getkey()
            if(s == 'p'):
                scene.forward = storedForward
                cont = False
        else:
            scene.forward = scene.forward*cos(.01) - sin(.01)*scene.up.cross(scene.forward)

''' places the info text on the screen '''
def placeInfoText():
    cameraControlInfo = label(pos=(.8,3.25,2.5), text='I, J, K, L to rotate camera.', height=40, border=2)
    orbitInfo = label(pos=(3,3.4,2.5), text='o to orbit', height=30)
    orbitInfoLn2 = label(pos=(2.9,3.15,2.5), text ='p to cancel orbit', height=30)
    controllerInfo = label(pos=(-10,-10,-2.5), text='q,w,e', height = 20)
    cI2 = label(pos=(-10.5,-12,-2.5), text='a,s,d', height=20)
    cI3 = label(pos=(-11,-14,-2.5), text='z,x,c', height=20)
    cI4 = label(pos=(-11.5,-16,-2.5), text='to select and move the cube', height=20)
    esc = label(pos=(3.5,1.5,2.5), text='esc to quit', height=30)
    ''' reset function does not work, so its label is excluded
    reset = label(pos=(3,3.65,2.5), text='r to reset', height=30) '''
    scramble = label(pos=(2,3.65,2.5), text='b to scramble', height=30)
    solve = label(pos=(.75,3.6,2.5), text='f to solve', height=30)
    undo = label(pos=(-.75,3.6,2.5), text='v to undo', height=30)
    useKeyboard = label(pos=(-2.9,2.1,0), text='press keys on keyboard', height=30)
    useKeyboard2 = label(pos=(-4,1.4,0), text='to control', height=30)

def txt(txt, npos):
    return text(text = txt, align = 'center', depth = -.3, color = color.black, pos = npos, axis = (1.5, -.05, -1))
             
    
''' debug print methods '''
def printCubeletPos(cubelet):
    print(cubelet.pos[0], cubelet.pos[1], cubelet.pos[2])
    
def printCubeletListPos(cubeletList):
    for i in range(len(cubeletList)):
        printCubeletPos(cubeletList[i])
    print("")

def printRCState(rCube):
    print("rCState")
    for i in range(3):
        for j in range(3):
            print(printCubeletPos(rCube.rCState[i][j][0]), printCubeletPos(rCube.rCState[i][j][1]), printCubeletPos(rCube.rCState[i][j][2]))
    print("")

def printRCCubeletList(rCube):
    print("cubeletList")
    for i in range(3):
        for j in range(3):
            print(printCubeletPos(rCube.cubeletList[i][j][0]), printCubeletPos(rCube.cubeletList[i][j][1]), printCubeletPos(rCube.cubeletList[i][j][2]))
    print("")
                
        
''' instantiates scene variables to make the app look better '''
def initScene():
    scene.x = 200
    scene.y = 200
    scene.width = 800
    scene.height = 800
    scene.title = "Robert Kaucic's VPython Rubik's Cube"
    scene.ambient = 1
    scene.background = (.5,.5,.5)
    scene.forward = (-1,-1,-1.5)
    scene.range = 4.5
    scene.userzoom = False
    scene.userspin = False

''' executed code '''
initScene()
placeInfoText()
rCube = RubiksCube()
cont = True
while(cont):
    cont = rCube.runOneMove()
exit()
